item = {'collectionIds': [], 'folderId': None, 'favorite': False, 'edit': True, 'viewPassword': True, 'permissions': {'delete': True, 'restore': True}, 'id': '3d9623ea-9fa3-4c57-85a2-b2c70140886e', 'organizationId': None, 'type': 1, 'data': {'uri': None, 'uris': None, 'username': '2.H+/zYRZppGBUeY+n3v+Msw==|Yb6DcEM2+t8swIUcZoFq2g==|qisUGIl3/9p4oyIQO0HpIQPSYWkdK0ZXIhSBqVXyBy8=', 'password': '2.6W7alo595LJoN80a/COGuw==|tABysM4+jmaxx3DJPgjxEQ==|wvHeEkgnx2C6s7EeLEC4+vcD403enPzJn7pf+4Tibfc=', 'passwordRevisionDate': None, 'totp': None, 'autofillOnPageLoad': None, 'fido2Credentials': None, 'name': '2.4Q4EK8nnwiCtE2hW+QXqPA==|iUGx0fwAMXHQS5ffv/GO8w==|kW1N8wyk98w68UgD1v3XkfZ8tKGnIGQDD4lPP9Qm2po=', 'notes': None, 'fields': None, 'passwordHistory': None}, 'name': '2.4Q4EK8nnwiCtE2hW+QXqPA==|iUGx0fwAMXHQS5ffv/GO8w==|kW1N8wyk98w68UgD1v3XkfZ8tKGnIGQDD4lPP9Qm2po=', 'notes': None, 'login': {'uri': None, 'uris': None, 'username': '2.H+/zYRZppGBUeY+n3v+Msw==|Yb6DcEM2+t8swIUcZoFq2g==|qisUGIl3/9p4oyIQO0HpIQPSYWkdK0ZXIhSBqVXyBy8=', 'password': '2.6W7alo595LJoN80a/COGuw==|tABysM4+jmaxx3DJPgjxEQ==|wvHeEkgnx2C6s7EeLEC4+vcD403enPzJn7pf+4Tibfc=', 'passwordRevisionDate': None, 'totp': None, 'autofillOnPageLoad': None, 'fido2Credentials': None}, 'card': None, 'identity': None, 'secureNote': None, 'sshKey': None, 'fields': None, 'passwordHistory': None, 'attachments': None, 'organizationUseTotp': False, 'revisionDate': '2025-04-22T19:27:01.48Z', 'creationDate': '2025-04-22T19:27:01.48Z', 'deletedDate': None, 'reprompt': 0, 'key': '2.HuqX3TN9v/6T37s+lVQ5lg==|83TAeYavnUyQAYuQOijpS2+AAQU3ObrwUcqW/cLFKDwmZBfAkmJ2QygA6f66SgIGa98vV7XUaYBzgm3uuKQpeG0o9mURBoM7wljkVt239yo=|1mRun2x/YNR6GxWyc91sMDFkocSbDovLYIqVPouL21M=', 'object': 'cipherDetails'}
